@@ -20,26 +20,67 @@
 #include <utils.h>
 
 struct netlink_linearize_ctx {
-	struct nft_rule		*nlr;
+	struct nftnl_rule	*nlr;
 	unsigned int		reg_low;
 };
 
-static void netlink_put_register(struct nft_rule_expr *nle,
+static void netlink_put_register(struct nftnl_expr *nle,
 				 uint32_t attr, uint32_t reg)
 {
-	nft_rule_expr_set_u32(nle, attr, reg);
+	/* Convert to 128 bit register numbers if possible for compatibility */
+	if (reg != NFT_REG_VERDICT) {
+		reg -= NFT_REG_1;
+		if (reg % (NFT_REG_SIZE / NFT_REG32_SIZE) == 0)
+			reg = NFT_REG_1 + reg / (NFT_REG_SIZE / NFT_REG32_SIZE);
+		else
+			reg += NFT_REG32_00;
+	}
+
+	nftnl_expr_set_u32(nle, attr, reg);
 }
 
-static enum nft_registers get_register(struct netlink_linearize_ctx *ctx)
+static enum nft_registers __get_register(struct netlink_linearize_ctx *ctx,
+					 unsigned int size)
 {
-	if (ctx->reg_low > NFT_REG_MAX)
+	unsigned int reg, n;
+
+	n = netlink_register_space(size);
+	if (ctx->reg_low + n > NFT_REG_1 + NFT_REG32_15 - NFT_REG32_00 + 1)
 		BUG("register reg_low %u invalid\n", ctx->reg_low);
-	return ctx->reg_low++;
+
+	reg = ctx->reg_low;
+	ctx->reg_low += n;
+	return reg;
 }
 
-static void release_register(struct netlink_linearize_ctx *ctx)
+static void __release_register(struct netlink_linearize_ctx *ctx,
+			       unsigned int size)
 {
-	ctx->reg_low--;
+	unsigned int n;
+
+	n = netlink_register_space(size);
+	if (ctx->reg_low < NFT_REG_1 + n)
+		BUG("register reg_low %u invalid\n", ctx->reg_low);
+
+	ctx->reg_low -= n;
+}
+
+static enum nft_registers get_register(struct netlink_linearize_ctx *ctx,
+				       const struct expr *expr)
+{
+	if (expr && expr->ops->type == EXPR_CONCAT)
+		return __get_register(ctx, expr->len);
+	else
+		return __get_register(ctx, NFT_REG_SIZE * BITS_PER_BYTE);
+}
+
+static void release_register(struct netlink_linearize_ctx *ctx,
+			     const struct expr *expr)
+{
+	if (expr && expr->ops->type == EXPR_CONCAT)
+		__release_register(ctx, expr->len);
+	else
+		__release_register(ctx, NFT_REG_SIZE * BITS_PER_BYTE);
 }
 
 static void netlink_gen_expr(struct netlink_linearize_ctx *ctx,
@@ -52,120 +93,122 @@ static void netlink_gen_concat(struct netlink_linearize_ctx *ctx,
 {
 	const struct expr *i;
 
-	list_for_each_entry(i, &expr->expressions, list)
+	list_for_each_entry(i, &expr->expressions, list) {
 		netlink_gen_expr(ctx, i, dreg);
+		dreg += netlink_register_space(i->len);
+	}
 }
 
 static void netlink_gen_payload(struct netlink_linearize_ctx *ctx,
 				const struct expr *expr,
 				enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("payload");
-	netlink_put_register(nle, NFT_EXPR_PAYLOAD_DREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_BASE,
+	netlink_put_register(nle, NFTNL_EXPR_PAYLOAD_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_PAYLOAD_BASE,
 			      expr->payload.base - 1);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_OFFSET,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_PAYLOAD_OFFSET,
 			      expr->payload.offset / BITS_PER_BYTE);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_PAYLOAD_LEN,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_PAYLOAD_LEN,
 			      expr->len / BITS_PER_BYTE);
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_exthdr(struct netlink_linearize_ctx *ctx,
 			       const struct expr *expr,
 			       enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("exthdr");
-	netlink_put_register(nle, NFT_EXPR_EXTHDR_DREG, dreg);
-	nft_rule_expr_set_u8(nle, NFT_EXPR_EXTHDR_TYPE,
+	netlink_put_register(nle, NFTNL_EXPR_EXTHDR_DREG, dreg);
+	nftnl_expr_set_u8(nle, NFTNL_EXPR_EXTHDR_TYPE,
 			     expr->exthdr.desc->type);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_EXTHDR_OFFSET,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_EXTHDR_OFFSET,
 			      expr->exthdr.tmpl->offset / BITS_PER_BYTE);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_EXTHDR_LEN,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_EXTHDR_LEN,
 			      expr->len / BITS_PER_BYTE);
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_meta(struct netlink_linearize_ctx *ctx,
 			     const struct expr *expr,
 			     enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("meta");
-	netlink_put_register(nle, NFT_EXPR_META_DREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_META_KEY, expr->meta.key);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_META_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_META_KEY, expr->meta.key);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_ct(struct netlink_linearize_ctx *ctx,
 			   const struct expr *expr,
 			   enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("ct");
-	netlink_put_register(nle, NFT_EXPR_CT_DREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_KEY, expr->ct.key);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_CT_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_CT_KEY, expr->ct.key);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_map(struct netlink_linearize_ctx *ctx,
 			    const struct expr *expr,
 			    enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 
 	assert(expr->mappings->ops->type == EXPR_SET_REF);
 
 	if (dreg == NFT_REG_VERDICT)
-		sreg = get_register(ctx);
+		sreg = get_register(ctx, expr->map);
 	else
 		sreg = dreg;
 
 	netlink_gen_expr(ctx, expr->map, sreg);
 
 	nle = alloc_nft_expr("lookup");
-	netlink_put_register(nle, NFT_EXPR_LOOKUP_SREG, sreg);
-	netlink_put_register(nle, NFT_EXPR_LOOKUP_DREG, dreg);
-	nft_rule_expr_set_str(nle, NFT_EXPR_LOOKUP_SET,
+	netlink_put_register(nle, NFTNL_EXPR_LOOKUP_SREG, sreg);
+	netlink_put_register(nle, NFTNL_EXPR_LOOKUP_DREG, dreg);
+	nftnl_expr_set_str(nle, NFTNL_EXPR_LOOKUP_SET,
 			      expr->mappings->set->handle.set);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SET_ID,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_LOOKUP_SET_ID,
 			      expr->mappings->set->handle.set_id);
 
 	if (dreg == NFT_REG_VERDICT)
-		release_register(ctx);
+		release_register(ctx, expr->map);
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_lookup(struct netlink_linearize_ctx *ctx,
 			       const struct expr *expr,
 			       enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 
 	assert(expr->right->ops->type == EXPR_SET_REF);
 	assert(dreg == NFT_REG_VERDICT);
 
-	sreg = get_register(ctx);
+	sreg = get_register(ctx, expr->left);
 	netlink_gen_expr(ctx, expr->left, sreg);
 
 	nle = alloc_nft_expr("lookup");
-	netlink_put_register(nle, NFT_EXPR_LOOKUP_SREG, sreg);
-	nft_rule_expr_set_str(nle, NFT_EXPR_LOOKUP_SET,
+	netlink_put_register(nle, NFTNL_EXPR_LOOKUP_SREG, sreg);
+	nftnl_expr_set_str(nle, NFTNL_EXPR_LOOKUP_SET,
 			      expr->right->set->handle.set);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_LOOKUP_SET_ID,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_LOOKUP_SET_ID,
 			      expr->right->set->handle.set_id);
 
-	release_register(ctx);
-	nft_rule_add_expr(ctx->nlr, nle);
+	release_register(ctx, expr->left);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static enum nft_cmp_ops netlink_gen_cmp_op(enum ops op)
@@ -196,14 +239,17 @@ static void netlink_gen_cmp(struct netlink_linearize_ctx *ctx,
 			    const struct expr *expr,
 			    enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 	struct nft_data_linearize nld, zero = {};
 	struct expr *right;
 
 	assert(dreg == NFT_REG_VERDICT);
 
-	sreg = get_register(ctx);
+	if (expr->right->ops->type == EXPR_RANGE)
+		return netlink_gen_range(ctx, expr, dreg);
+
+	sreg = get_register(ctx, expr->left);
 	netlink_gen_expr(ctx, expr->left, sreg);
 
 	switch (expr->right->ops->type) {
@@ -219,31 +265,29 @@ static void netlink_gen_cmp(struct netlink_linearize_ctx *ctx,
 		zero.len = nld.len;
 
 		nle = alloc_nft_expr("bitwise");
-		netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, sreg);
-		netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, sreg);
-		nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, nld.len);
-		nft_rule_expr_set(nle, NFT_EXPR_BITWISE_MASK, &nld.value, nld.len);
-		nft_rule_expr_set(nle, NFT_EXPR_BITWISE_XOR, &zero.value, zero.len);
-		nft_rule_add_expr(ctx->nlr, nle);
+		netlink_put_register(nle, NFTNL_EXPR_BITWISE_SREG, sreg);
+		netlink_put_register(nle, NFTNL_EXPR_BITWISE_DREG, sreg);
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_LEN, nld.len);
+		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, &nld.value, nld.len);
+		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, &zero.value, zero.len);
+		nftnl_rule_add_expr(ctx->nlr, nle);
 
 		right = expr->right->prefix;
 		break;
 		}
-	case EXPR_RANGE:
-		return netlink_gen_range(ctx, expr, dreg);
 	default:
 		right = expr->right;
 	}
 
 	nle = alloc_nft_expr("cmp");
-	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
+	netlink_put_register(nle, NFTNL_EXPR_CMP_SREG, sreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP,
 			      netlink_gen_cmp_op(expr->op));
 	netlink_gen_data(right, &nld);
-	nft_rule_expr_set(nle, NFT_EXPR_CMP_DATA, nld.value, nld.len);
-	release_register(ctx);
+	nftnl_expr_set(nle, NFTNL_EXPR_CMP_DATA, nld.value, nld.len);
+	release_register(ctx, expr->left);
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
@@ -251,25 +295,25 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 			      enum nft_registers dreg)
 {
 	struct expr *range = expr->right;
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 	struct nft_data_linearize nld;
 
 	assert(dreg == NFT_REG_VERDICT);
 
-	sreg = get_register(ctx);
+	sreg = get_register(ctx, expr->left);
 	netlink_gen_expr(ctx, expr->left, sreg);
 
 	nle = alloc_nft_expr("cmp");
-	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFTNL_EXPR_CMP_SREG, sreg);
 	switch (expr->op) {
 	case OP_NEQ:
-		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP,
 				      netlink_gen_cmp_op(OP_LT));
 		break;
 	case OP_RANGE:
 	case OP_EQ:
-		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP,
 				      netlink_gen_cmp_op(OP_GTE));
 		break;
 	default:
@@ -277,19 +321,19 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 	}
 
 	netlink_gen_data(range->left, &nld);
-	nft_rule_expr_set(nle, NFT_EXPR_CMP_DATA, nld.value, nld.len);
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_expr_set(nle, NFTNL_EXPR_CMP_DATA, nld.value, nld.len);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 
 	nle = alloc_nft_expr("cmp");
-	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
+	netlink_put_register(nle, NFTNL_EXPR_CMP_SREG, sreg);
 	switch (expr->op) {
 	case OP_NEQ:
-		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP,
 				      netlink_gen_cmp_op(OP_GT));
 		break;
 	case OP_RANGE:
 	case OP_EQ:
-		nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP,
 				      netlink_gen_cmp_op(OP_LTE));
 		break;
 	default:
@@ -297,17 +341,17 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 	}
 
 	netlink_gen_data(range->right, &nld);
-	nft_rule_expr_set(nle, NFT_EXPR_CMP_DATA, nld.value, nld.len);
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_expr_set(nle, NFTNL_EXPR_CMP_DATA, nld.value, nld.len);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 
-	release_register(ctx);
+	release_register(ctx, expr->left);
 }
 
 static void netlink_gen_flagcmp(struct netlink_linearize_ctx *ctx,
 				const struct expr *expr,
 				enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	struct nft_data_linearize nld, nld2;
 	enum nft_registers sreg;
 	unsigned int len;
@@ -315,7 +359,7 @@ static void netlink_gen_flagcmp(struct netlink_linearize_ctx *ctx,
 
 	assert(dreg == NFT_REG_VERDICT);
 
-	sreg = get_register(ctx);
+	sreg = get_register(ctx, expr->left);
 	netlink_gen_expr(ctx, expr->left, sreg);
 	len = div_round_up(expr->left->len, BITS_PER_BYTE);
 
@@ -325,21 +369,21 @@ static void netlink_gen_flagcmp(struct netlink_linearize_ctx *ctx,
 	netlink_gen_data(expr->right, &nld2);
 
 	nle = alloc_nft_expr("bitwise");
-	netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, sreg);
-	netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, len);
-	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_MASK, &nld2.value, nld2.len);
-	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_XOR, &nld.value, nld.len);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_BITWISE_SREG, sreg);
+	netlink_put_register(nle, NFTNL_EXPR_BITWISE_DREG, sreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_LEN, len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, &nld2.value, nld2.len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, &nld.value, nld.len);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 
 	nle = alloc_nft_expr("cmp");
-	netlink_put_register(nle, NFT_EXPR_CMP_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CMP_OP, NFT_CMP_NEQ);
-	nft_rule_expr_set(nle, NFT_EXPR_CMP_DATA, nld.value, nld.len);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_CMP_SREG, sreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_CMP_OP, NFT_CMP_NEQ);
+	nftnl_expr_set(nle, NFTNL_EXPR_CMP_DATA, nld.value, nld.len);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 
 	mpz_clear(zero);
-	release_register(ctx);
+	release_register(ctx, expr->left);
 }
 
 static void netlink_gen_relational(struct netlink_linearize_ctx *ctx,
@@ -378,7 +422,7 @@ static void netlink_gen_binop(struct netlink_linearize_ctx *ctx,
 			      const struct expr *expr,
 			      enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	struct nft_data_linearize nld;
 	struct expr *left, *i;
 	struct expr *binops[16];
@@ -425,21 +469,21 @@ static void netlink_gen_binop(struct netlink_linearize_ctx *ctx,
 	len = div_round_up(expr->len, BITS_PER_BYTE);
 
 	nle = alloc_nft_expr("bitwise");
-	netlink_put_register(nle, NFT_EXPR_BITWISE_SREG, dreg);
-	netlink_put_register(nle, NFT_EXPR_BITWISE_DREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BITWISE_LEN, len);
+	netlink_put_register(nle, NFTNL_EXPR_BITWISE_SREG, dreg);
+	netlink_put_register(nle, NFTNL_EXPR_BITWISE_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_LEN, len);
 
 	netlink_gen_raw_data(mask, expr->byteorder, len, &nld);
-	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_MASK, nld.value, nld.len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, nld.value, nld.len);
 	netlink_gen_raw_data(xor, expr->byteorder, len, &nld);
-	nft_rule_expr_set(nle, NFT_EXPR_BITWISE_XOR, nld.value, nld.len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, nld.value, nld.len);
 
 	mpz_clear(tmp);
 	mpz_clear(val);
 	mpz_clear(xor);
 	mpz_clear(mask);
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static enum nft_byteorder_ops netlink_gen_unary_op(enum ops op)
@@ -458,54 +502,56 @@ static void netlink_gen_unary(struct netlink_linearize_ctx *ctx,
 			      const struct expr *expr,
 			      enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	netlink_gen_expr(ctx, expr->arg, dreg);
 
 	nle = alloc_nft_expr("byteorder");
-	netlink_put_register(nle, NFT_EXPR_BYTEORDER_SREG, dreg);
-	netlink_put_register(nle, NFT_EXPR_BYTEORDER_DREG, dreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_LEN,
+	netlink_put_register(nle, NFTNL_EXPR_BYTEORDER_SREG, dreg);
+	netlink_put_register(nle, NFTNL_EXPR_BYTEORDER_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_BYTEORDER_LEN,
 			      expr->len / BITS_PER_BYTE);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_SIZE,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_BYTEORDER_SIZE,
 			      expr->arg->len % 32 ? 2 : 4);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_BYTEORDER_OP,
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_BYTEORDER_OP,
 			      netlink_gen_unary_op(expr->op));
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_immediate(struct netlink_linearize_ctx *ctx,
 				  const struct expr *expr,
 				  enum nft_registers dreg)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	struct nft_data_linearize nld;
 
 	nle = alloc_nft_expr("immediate");
-	netlink_put_register(nle, NFT_EXPR_IMM_DREG, dreg);
+	netlink_put_register(nle, NFTNL_EXPR_IMM_DREG, dreg);
 	netlink_gen_data(expr, &nld);
 	switch (expr->ops->type) {
 	case EXPR_VALUE:
-		nft_rule_expr_set(nle, NFT_EXPR_IMM_DATA, nld.value, nld.len);
+		nftnl_expr_set(nle, NFTNL_EXPR_IMM_DATA, nld.value, nld.len);
 		break;
 	case EXPR_VERDICT:
 		if ((expr->chain != NULL) &&
-		    !nft_rule_expr_is_set(nle, NFT_EXPR_IMM_CHAIN)) {
-			nft_rule_expr_set_str(nle, NFT_EXPR_IMM_CHAIN,
+		    !nftnl_expr_is_set(nle, NFTNL_EXPR_IMM_CHAIN)) {
+			nftnl_expr_set_str(nle, NFTNL_EXPR_IMM_CHAIN,
 					      nld.chain);
 		}
-		nft_rule_expr_set_u32(nle, NFT_EXPR_IMM_VERDICT, nld.verdict);
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_IMM_VERDICT, nld.verdict);
 		break;
 	default:
 		break;
 	}
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_expr(struct netlink_linearize_ctx *ctx,
 			     const struct expr *expr,
 			     enum nft_registers dreg)
 {
+	assert(dreg < ctx->reg_low);
+
 	switch (expr->ops->type) {
 	case EXPR_VERDICT:
 	case EXPR_VALUE:
@@ -528,6 +574,8 @@ static void netlink_gen_expr(struct netlink_linearize_ctx *ctx,
 		return netlink_gen_meta(ctx, expr, dreg);
 	case EXPR_CT:
 		return netlink_gen_ct(ctx, expr, dreg);
+	case EXPR_SET_ELEM:
+		return netlink_gen_expr(ctx, expr->key, dreg);
 	default:
 		BUG("unknown expression type %s\n", expr->ops->name);
 	}
@@ -542,222 +590,222 @@ static void netlink_gen_verdict_stmt(struct netlink_linearize_ctx *ctx,
 static void netlink_gen_counter_stmt(struct netlink_linearize_ctx *ctx,
 				     const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("counter");
 	if (stmt->counter.packets) {
-		nft_rule_expr_set_u64(nle, NFT_EXPR_CTR_PACKETS,
+		nftnl_expr_set_u64(nle, NFTNL_EXPR_CTR_PACKETS,
 				      stmt->counter.packets);
 	}
 	if (stmt->counter.bytes) {
-		nft_rule_expr_set_u64(nle, NFT_EXPR_CTR_BYTES,
+		nftnl_expr_set_u64(nle, NFTNL_EXPR_CTR_BYTES,
 				      stmt->counter.bytes);
 	}
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_meta_stmt(struct netlink_linearize_ctx *ctx,
 				  const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 
-	sreg = get_register(ctx);
+	sreg = get_register(ctx, stmt->meta.expr);
 	netlink_gen_expr(ctx, stmt->meta.expr, sreg);
-	release_register(ctx);
+	release_register(ctx, stmt->meta.expr);
 
 	nle = alloc_nft_expr("meta");
-	netlink_put_register(nle, NFT_EXPR_META_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_META_KEY, stmt->meta.key);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_META_SREG, sreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_META_KEY, stmt->meta.key);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_log_stmt(struct netlink_linearize_ctx *ctx,
 				 const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("log");
 	if (stmt->log.prefix != NULL) {
-		nft_rule_expr_set_str(nle, NFT_EXPR_LOG_PREFIX,
+		nftnl_expr_set_str(nle, NFTNL_EXPR_LOG_PREFIX,
 				      stmt->log.prefix);
 	}
 	if (stmt->log.flags & STMT_LOG_GROUP) {
-		nft_rule_expr_set_u16(nle, NFT_EXPR_LOG_GROUP,
+		nftnl_expr_set_u16(nle, NFTNL_EXPR_LOG_GROUP,
 				      stmt->log.group);
 		if (stmt->log.flags & STMT_LOG_SNAPLEN)
-			nft_rule_expr_set_u32(nle, NFT_EXPR_LOG_SNAPLEN,
+			nftnl_expr_set_u32(nle, NFTNL_EXPR_LOG_SNAPLEN,
 					      stmt->log.snaplen);
 		if (stmt->log.flags & STMT_LOG_QTHRESHOLD)
-			nft_rule_expr_set_u16(nle, NFT_EXPR_LOG_QTHRESHOLD,
+			nftnl_expr_set_u16(nle, NFTNL_EXPR_LOG_QTHRESHOLD,
 					      stmt->log.qthreshold);
 	} else {
 		if (stmt->log.flags & STMT_LOG_LEVEL) {
-			nft_rule_expr_set_u32(nle, NFT_EXPR_LOG_LEVEL,
+			nftnl_expr_set_u32(nle, NFTNL_EXPR_LOG_LEVEL,
 					      stmt->log.level);
 		}
 	}
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_limit_stmt(struct netlink_linearize_ctx *ctx,
 				   const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("limit");
-	nft_rule_expr_set_u64(nle, NFT_EXPR_LIMIT_RATE, stmt->limit.rate);
-	nft_rule_expr_set_u64(nle, NFT_EXPR_LIMIT_UNIT, stmt->limit.unit);
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_expr_set_u64(nle, NFTNL_EXPR_LIMIT_RATE, stmt->limit.rate);
+	nftnl_expr_set_u64(nle, NFTNL_EXPR_LIMIT_UNIT, stmt->limit.unit);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_reject_stmt(struct netlink_linearize_ctx *ctx,
 				    const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("reject");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_REJECT_TYPE, stmt->reject.type);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_REJECT_TYPE, stmt->reject.type);
 	if (stmt->reject.icmp_code != -1)
-		nft_rule_expr_set_u8(nle, NFT_EXPR_REJECT_CODE,
+		nftnl_expr_set_u8(nle, NFTNL_EXPR_REJECT_CODE,
 				     stmt->reject.icmp_code);
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_nat_stmt(struct netlink_linearize_ctx *ctx,
 				 const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers amin_reg, amax_reg;
 	enum nft_registers pmin_reg, pmax_reg;
 	int registers = 0;
 	int family;
 
 	nle = alloc_nft_expr("nat");
-	nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_TYPE, stmt->nat.type);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_NAT_TYPE, stmt->nat.type);
 
-	family = nft_rule_attr_get_u32(ctx->nlr, NFT_RULE_ATTR_FAMILY);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_FAMILY, family);
+	family = nftnl_rule_get_u32(ctx->nlr, NFTNL_RULE_FAMILY);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_NAT_FAMILY, family);
 
 	if (stmt->nat.flags != 0)
-		nft_rule_expr_set_u32(nle, NFT_EXPR_NAT_FLAGS, stmt->nat.flags);
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_NAT_FLAGS, stmt->nat.flags);
 
 	if (stmt->nat.addr) {
-		amin_reg = get_register(ctx);
+		amin_reg = get_register(ctx, NULL);
 		registers++;
 
 		if (stmt->nat.addr->ops->type == EXPR_RANGE) {
-			amax_reg = get_register(ctx);
+			amax_reg = get_register(ctx, NULL);
 			registers++;
 
 			netlink_gen_expr(ctx, stmt->nat.addr->left, amin_reg);
 			netlink_gen_expr(ctx, stmt->nat.addr->right, amax_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_ADDR_MIN,
 					     amin_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MAX,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_ADDR_MAX,
 					     amax_reg);
 		} else {
 			netlink_gen_expr(ctx, stmt->nat.addr, amin_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_ADDR_MIN,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_ADDR_MIN,
 					     amin_reg);
 		}
 
 	}
 
 	if (stmt->nat.proto) {
-		pmin_reg = get_register(ctx);
+		pmin_reg = get_register(ctx, NULL);
 		registers++;
 
 		if (stmt->nat.proto->ops->type == EXPR_RANGE) {
-			pmax_reg = get_register(ctx);
+			pmax_reg = get_register(ctx, NULL);
 			registers++;
 
 			netlink_gen_expr(ctx, stmt->nat.proto->left, pmin_reg);
 			netlink_gen_expr(ctx, stmt->nat.proto->right, pmax_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_PROTO_MIN,
 					     pmin_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MAX,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_PROTO_MAX,
 					     pmax_reg);
 		} else {
 			netlink_gen_expr(ctx, stmt->nat.proto, pmin_reg);
-			netlink_put_register(nle, NFT_EXPR_NAT_REG_PROTO_MIN,
+			netlink_put_register(nle, NFTNL_EXPR_NAT_REG_PROTO_MIN,
 					     pmin_reg);
 		}
 	}
 
 	while (registers > 0) {
-		release_register(ctx);
+		release_register(ctx, NULL);
 		registers--;
 	}
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_masq_stmt(struct netlink_linearize_ctx *ctx,
 				  const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 
 	nle = alloc_nft_expr("masq");
 	if (stmt->masq.flags != 0)
-		nft_rule_expr_set_u32(nle, NFT_EXPR_MASQ_FLAGS,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_MASQ_FLAGS,
 				      stmt->masq.flags);
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_redir_stmt(struct netlink_linearize_ctx *ctx,
 				   const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers pmin_reg, pmax_reg;
 	int registers = 0;
 
 	nle = alloc_nft_expr("redir");
 
 	if (stmt->redir.flags != 0)
-		nft_rule_expr_set_u32(nle, NFT_EXPR_REDIR_FLAGS,
+		nftnl_expr_set_u32(nle, NFTNL_EXPR_REDIR_FLAGS,
 				      stmt->redir.flags);
 
 	if (stmt->redir.proto) {
-		pmin_reg = get_register(ctx);
+		pmin_reg = get_register(ctx, NULL);
 		registers++;
 
 		if (stmt->redir.proto->ops->type == EXPR_RANGE) {
-			pmax_reg = get_register(ctx);
+			pmax_reg = get_register(ctx, NULL);
 			registers++;
 
 			netlink_gen_expr(ctx, stmt->redir.proto->left,
 					 pmin_reg);
 			netlink_gen_expr(ctx, stmt->redir.proto->right,
 					 pmax_reg);
-			nft_rule_expr_set_u32(nle,
-					      NFT_EXPR_REDIR_REG_PROTO_MIN,
+			nftnl_expr_set_u32(nle,
+					      NFTNL_EXPR_REDIR_REG_PROTO_MIN,
 					      pmin_reg);
-			nft_rule_expr_set_u32(nle,
-					      NFT_EXPR_REDIR_REG_PROTO_MAX,
+			nftnl_expr_set_u32(nle,
+					      NFTNL_EXPR_REDIR_REG_PROTO_MAX,
 					      pmax_reg);
 		} else {
 			netlink_gen_expr(ctx, stmt->redir.proto, pmin_reg);
-			nft_rule_expr_set_u32(nle,
-					      NFT_EXPR_REDIR_REG_PROTO_MIN,
+			nftnl_expr_set_u32(nle,
+					      NFTNL_EXPR_REDIR_REG_PROTO_MIN,
 					      pmin_reg);
 		}
 	}
 
 	while (registers > 0) {
-		release_register(ctx);
+		release_register(ctx, NULL);
 		registers--;
 	}
 
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_queue_stmt(struct netlink_linearize_ctx *ctx,
 				 const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	uint16_t total_queues;
 	mpz_t low, high;
 
@@ -770,13 +818,13 @@ static void netlink_gen_queue_stmt(struct netlink_linearize_ctx *ctx,
 	total_queues = mpz_get_uint16(high) - mpz_get_uint16(low) + 1;
 
 	nle = alloc_nft_expr("queue");
-	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_NUM, mpz_get_uint16(low));
-	nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_TOTAL, total_queues);
+	nftnl_expr_set_u16(nle, NFTNL_EXPR_QUEUE_NUM, mpz_get_uint16(low));
+	nftnl_expr_set_u16(nle, NFTNL_EXPR_QUEUE_TOTAL, total_queues);
 	if (stmt->queue.flags) {
-		nft_rule_expr_set_u16(nle, NFT_EXPR_QUEUE_FLAGS,
+		nftnl_expr_set_u16(nle, NFTNL_EXPR_QUEUE_FLAGS,
 				      stmt->queue.flags);
 	}
-	nft_rule_add_expr(ctx->nlr, nle);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 
 	mpz_clear(low);
 	mpz_clear(high);
@@ -785,17 +833,39 @@ static void netlink_gen_queue_stmt(struct netlink_linearize_ctx *ctx,
 static void netlink_gen_ct_stmt(struct netlink_linearize_ctx *ctx,
 				  const struct stmt *stmt)
 {
-	struct nft_rule_expr *nle;
+	struct nftnl_expr *nle;
 	enum nft_registers sreg;
 
-	sreg = get_register(ctx);
+	sreg = get_register(ctx, stmt->ct.expr);
 	netlink_gen_expr(ctx, stmt->ct.expr, sreg);
-	release_register(ctx);
+	release_register(ctx, stmt->ct.expr);
 
 	nle = alloc_nft_expr("ct");
-	netlink_put_register(nle, NFT_EXPR_CT_SREG, sreg);
-	nft_rule_expr_set_u32(nle, NFT_EXPR_CT_KEY, stmt->ct.key);
-	nft_rule_add_expr(ctx->nlr, nle);
+	netlink_put_register(nle, NFTNL_EXPR_CT_SREG, sreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_CT_KEY, stmt->ct.key);
+	nftnl_rule_add_expr(ctx->nlr, nle);
+}
+
+static void netlink_gen_set_stmt(struct netlink_linearize_ctx *ctx,
+				 const struct stmt *stmt)
+{
+	struct nftnl_expr *nle;
+	enum nft_registers sreg_key;
+
+	sreg_key = get_register(ctx, stmt->set.key);
+	netlink_gen_expr(ctx, stmt->set.key, sreg_key);
+	release_register(ctx, stmt->set.key);
+
+	nle = alloc_nft_expr("dynset");
+	netlink_put_register(nle, NFTNL_EXPR_DYNSET_SREG_KEY, sreg_key);
+	nftnl_expr_set_u64(nle, NFTNL_EXPR_DYNSET_TIMEOUT,
+			      stmt->set.key->timeout);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_DYNSET_OP, stmt->set.op);
+	nftnl_expr_set_str(nle, NFTNL_EXPR_DYNSET_SET_NAME,
+			      stmt->set.set->set->handle.set);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_DYNSET_SET_ID,
+			      stmt->set.set->set->handle.set_id);
+	nftnl_rule_add_expr(ctx->nlr, nle);
 }
 
 static void netlink_gen_stmt(struct netlink_linearize_ctx *ctx,
@@ -826,12 +896,14 @@ static void netlink_gen_stmt(struct netlink_linearize_ctx *ctx,
 		return netlink_gen_queue_stmt(ctx, stmt);
 	case STMT_CT:
 		return netlink_gen_ct_stmt(ctx, stmt);
+	case STMT_SET:
+		return netlink_gen_set_stmt(ctx, stmt);
 	default:
 		BUG("unknown statement type %s\n", stmt->ops->name);
 	}
 }
 
-void netlink_linearize_rule(struct netlink_ctx *ctx, struct nft_rule *nlr,
+void netlink_linearize_rule(struct netlink_ctx *ctx, struct nftnl_rule *nlr,
 			    const struct rule *rule)
 {
 	struct netlink_linearize_ctx lctx;
